"""Play the Game of Nim, interactively or using the API


About
=====

This package implements the game called Nim. The game can be played
interactively in the CLI or in a web browser. You can install this package and
play the game. Easy::

> pip install nimgame
> nimgame

You can also try the game before installing the package, see the "TryIt" link
on the left (when it will work :-))

Those, who are more of the developer types, can use the API, after installing
the package. The game can then be embedded in other applications and played
from any other kind of a GUI.

For mathematicians/statisticians, it can also be used for creating statistics
by running multiple games. If you know `combinatorial game theory`_ terms like
`sequential game`_, `perfect information`_, `impartial game`_,
`Sprague–Grundy theorem`_ and such, this API can be useful for you.


Game Rules
===========

For the general description of the Nim and similar games, see the
`Wikipedia page`_.

This game is played by 2 palyers (the computer can be a player).
There are coins arranged in heaps. Some descriptions call the grouped objects
stones, shells, matches, and the grouping may also be called piles.
There may be any number of heaps and any number of coins in a heap.

Players take turns. In each turn, a player takes coins from a heap (only 1
heap). It must be at least 1 coin. There is no upper limit (but the heap size).

There are 2 types of the game:

- the "normal" game is where the player, who takes the last coin, wins
- the "misère_" game is where the player, who has to take the last coin, loses


Usage
======

After instantiating the Nim class, you need to get the heaps set up. The
starting heaps then can be analysed and the starting player set. If it is the
Computer, it automatically does the 1\ :sup:`st` move. Then the Player is to do
the next move, then the Computer moves, and so on. When no more coins left, the
game ends.

.. code:: python

    import nimgame
    nim = nimgame.Nim(error_rate=10)  # Make the Computer do 10% errors
    nim.setup_heaps()  # Create heaps randomly
    nim.set_start(myturn=True)  # Indicate that the Player wants to start
    nim.do_move(nim.Move('b', 4))  # From heap 'B', remove 4 coins
    ...
    if nim.game_end():
        exit(0)


Available objects in the package
================================

While different objects are defined in the source at different levels (e.g. the
Nim class in the source/core.py module), the necessary ones are made available
at the package level, i.e. after importing nimgame, the `nimgame.Nim` refers to
that class.

Nim
    The main class of the game
play_CLI
    A function that interactively plays the game in CLI
Move
    A :py:func:`collections.namedtuple` with `heapdesig` and `removecount`
    fields for defining a move, i.e. what heap is to be reduced and by how many
    coins.
    `heapdesig` can be an int (heap numbers start from zero), or a letter (case
    insensitive). E.g. number 1 is the same as "B" or "b".
ErrorRate
    A :py:func:`collections.namedtuple` with `Computer` and `Player` fields for
    defining the required error rates percentage for the players. Only used when
    both of them are simulated.
HeapCoinRange
    A :py:func:`collections.namedtuple` with `min` and `max` fields for defining
    the ranges for numbers of heaps and heap sizes. These ranges are used for
    automatic heap creation.


Package content
===============

:mod:`core`
    It provides the :class:`core.Nim` class and its public methods

:mod:`calculations`
    A mixin of the :class:`core.Nim` class with calculation methods

:mod:`playing.demo`
    Functions to play the game. It can be run to do interactions for executing
    automatic runs or interactive games. The same is executed when
    
    -   the package itself is called (see the "__main__.py" module), e.g.
    
        :command:`python -m nimgame`
        
    -   the :command:`nimgame.exe` (in Scripts) is executed.
        
        NB, the python Scripts dir is usually in the execution path, i.e. the
        nimgame.exe should be available from anywhere in the CLI prompt.

:mod:`tests.testruns`
    You can test the function by running several games automatically and gather
    statistics.


.. _Wikipedia page: https://en.wikipedia.org/wiki/Nim
.. _misère: https://en.wikipedia.org/wiki/Mis%C3%A8re#Mis%C3%A8re_game
.. _combinatorial game theory: https://en.wikipedia.org/wiki/Combinatorial_game_theory
.. _sequential game: https://en.wikipedia.org/wiki/Sequential_game
.. _perfect information: https://en.wikipedia.org/wiki/Perfect_information
.. _impartial game: https://en.wikipedia.org/wiki/Impartial_game
.. _Sprague–Grundy theorem: https://en.wikipedia.org/wiki/Sprague%E2%80%93Grundy_theorem
"""
# Make sure the package base path is in the module import path (e.g. for flit)
import sys, pathlib
basepath = str(pathlib.PurePath(__file__).parent)
if basepath not in sys.path:
    sys.path.insert(0, basepath)

# Version info to be made available globally
from source import version
__version_info__ = version.__version_info__
__version__ = version.__version__

# Make the Nim class available on package level
from source.core import Nim
# Make structures available on package level
from source.typedefs import Move, ErrorRate, HeapCoinRange

# Make the playing functions available on package level
from source.playing.cli.play import play_CLI
from source.playing.web.play import playweb

# For using the API in command line (e.g. Jupyter) and import all
__all__ = (
    'Nim', 
    'Move', 'ErrorRate', 'HeapCoinRange', 
    'play_CLI', 
    'playweb'
)



"""Dummy docstring"""



def ptgdemo():
    with ptg.WindowManager() as manager:
       demo = ptg.Window(
          ptg.Label("[210 bold]Hello world!"),
          ptg.Label(),
          ptg.InputField(prompt="Who are you?"),
          ptg.Label(),
          ptg.Button("Submit!")
       )
       
       manager.add(demo)
       manager.run()


if __name__ == '__main__':
    # An empty command seems to make the Windows cmd terminal capable of ANSI.
    # Note that importing pytermgui seems to emit an ESC[14t (setting maximum
    # number of lines in window to 14). In order to get this interpreted, the
    # empty command must be done BEFORE importing pytermgui
    import os; os.system('')

    import pytermgui as ptg
    
    ptgdemo()



"""Run tests of the Nim Game automatically

:func:`play_full_game` plays one full game automatically, doing the moves for
both parties, according to their set error rates in the :class:`Nim` instance

:func:`run_many_games` creates a :class:`Nim` instance and runs several games
automatically
"""


from source.core import Nim
from source.typedefs import ErrorRate, ErrorRate_T


def play_full_game(
        nim: Nim
) -> dict:
    """Play one full game
    
    There is no interaction, both players' moves are calculated by the program.
    We just call the 1st player as Computer and the 2nd as Player.
    
    It creates random heaps first. Then, figures out whether Computer is to
    start.
    This decision is based on the starting nimsum on Computer's favour, but with
    the Computer's error rate. Player and Computer take turns until the game
    ends. Moves are calculated with the same algorithm for the Computer and
    Player, but with different error rates.
    
    Args:
        nim: the :class:`Nim` instance, created with the init arguments, like
            `error_rate`

    Returns:
        The result of the game.
        Keys in the returned dict:
        
            - step_records: a list of statuses at each move
            - computer_won: a bool stating whether the "Computer" won
    """
    #create new random heaps
    nim.setup_heaps()
    #init compturn automatically
    nim.set_start()

    #init the step storage
    step_records = []
    step_records.append(
        nim.get_heapstatus() + f' before {nim.activeplayer}'
    )

    #do moves in a loop, until game end
    while not nim.game_end():
        #make a good decision?
        if nim.make_good_choice():
            #figure out the best move (what heap and how many coins)
            move = nim.figure_out_best_move()
        else:
            #set a random (most probably not good) move
            move = nim.get_random_move()

        #do and record a move
        nim.do_move(move)
        step_records.append(
            nim.get_heapstatus() + f' before {nim.activeplayer}'
        )

    #Computer won if it was its turn after the last coin was taken
    step_records.append(f'{nim.activeplayer} won')

    return dict(
        step_records=step_records,
        computer_won=(nim.activeplayer=='Computer') ^ (not nim.misere)
    )


def run_many_games(
    gamenum: int, 
    error_rate: ErrorRate_T
) -> None:
    """Run several games automatically
    
    This function starts with creating a :class:`Nim` instance with setting the
    required error rates. It then calls :func:`play_full_game` for each game. It
    prints a + when the "Computer" wins, and - when the "Player" wins.
    
    Then it prints the percentage the "Computer" won.
    
    Args:
        gamenum: the number of games to run
        error_rate: the required rate of bad decisions to be made by each player
    """
    #Computer's success
    won_games = 0

    #Computer failure details, in case it's needed
    lost_games = []

    #instantiate Nim
    nim = Nim(
        error_rate=error_rate, 
    )

    #do several games to collect statistics
    for _ in range(gamenum):
        #play one game
        game_result = play_full_game(nim)
        #Computer win?
        if game_result['computer_won']:
            won_games += 1
            print('+',  end='')
        else:
            lost_games.append(game_result['step_records'])
            print('-',  end='')
    print("\n")

    if won_games == gamenum:
        print(f"Computer won all the {gamenum} games, Player had no chance")
    else:
        if isinstance(error_rate, ErrorRate):
            error_rate = error_rate.Computer
        if error_rate:
            print(f"Computer won {won_games/gamenum*100}%")
        else:
            print(
                f"Computer only won {won_games/gamenum*100}%, "
                "in spite its error rate was 0%, WTF?!")
            for lost_game in lost_games:
                print('-'*40)
                for move in lost_game:
                    print(move)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# Insert the package path first, so that the package __init__ is found first
# before the source/__init__ module
import sys, pathlib
sys.path.insert(0, str(pathlib.Path('..').resolve()))
sys.path.insert(1, str(pathlib.Path('../source').resolve()))

# -- Project information -----------------------------------------------------

project = 'nimgame'
copyright = '2022, netcreator'
author = 'netcreator'

# The full version, major, minor, patch (taken from the source/version.py)
import version as ver
release = ver.__version__


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.napoleon', 
    'sphinx.ext.autodoc', 
    'sphinx.ext.autosummary', 
    'sphinx.ext.viewcode', 
    'sphinx.ext.intersphinx', 
]

napoleon_preprocess_types = True

autosummary_generate = False

intersphinx_mapping = {
    'python': ('https://docs.python.org/3.5', None)
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'alabaster'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Source Code display width correction
html_css_files = [
    'srccode_width_correction.css',
]



"""Demonstrate the features of the package with a small interactive execution

Since this module may be called from anywhere and even the package might not
have been installed yet, the package path is inserted into the sys.path if
missing. From that point on, imports like source.core work fine.

In order to demonstrate the correct usage of the package objects (e.g. the Nim
class), i.e. referring to it as `nimgame.Nim`, regardless of where this class is
actually created in the source, we create simulated namespace, fill it up with
the objects imported from their actual place, and call it nimgame.
"""


import types

# Insert the package into the system path
import sys, os
basepath = os.path.dirname(__file__)
packagepath = os.path.abspath(os.path.join(basepath, '../..'))
if packagepath not in sys.path:
    sys.path.insert(0, packagepath)

# Simulate "import nimgame" (as the package may not be installed yet)
from source.core import Nim as _Nim
from source.typedefs import ErrorRate as _ErrorRate
from source.playing.cli.play import play_CLI as _play_CLI
from source.playing.web.play import playweb as _playweb
nimgame = types.SimpleNamespace()
nimgame.Nim = _Nim
nimgame.ErrorRate = _ErrorRate
nimgame.play_CLI = _play_CLI
nimgame.playweb = _playweb

from source.playing.cli import ansi


def run():
    """Run the interactive CLI demo"""
    try:
        while True:
            testtype = input(
                'Press '
                '"m" for running multiple games'
                '; '
                '"p" for playing an interactive game'
                '; '
                '"w" for running the web server'
                '; [p]: '
            ) or 'p'
            ansi.clean_prev_line()

            if testtype=='m':
                #do the test with 1000 games
                gamecount = 1000
                #request 10% Computer and 20% Player error rate
                error_rate = nimgame.ErrorRate(Computer=10, Player=20)
                #run all tests
                from tests import testruns
                testruns.run_many_games(gamecount, error_rate)
                break
            
            elif testtype=='p':
                nimgame.play_CLI()
                break
            
            elif testtype=='w':
                nimgame.playweb()
                break

    except KeyboardInterrupt:
        #Ctrl-C happens when input is waiting, so delete that prompt
        ansi.clean_this_line()
        print('Game terminated by the user pressing Ctrl-C')
        exit(2)
        
    except Exception as e:
        print(e)
        exit(1)

    exit(0)


if __name__ == '__main__':
    run()



"""The Nim Game core

This main class implements the init and maintenance of heaps, movements of
coins, records the statuses and calculates decisions.
"""


from dataclasses import dataclass
import itertools
import random; random.seed()
import valid8 as v8
from source.typedefs import Move, ErrorRate, ErrorRate_T, ErrorRateTypes, \
    MyTurn_T, HeapCoinRange, HeapCoinRange_T, HeapCoinRangeTypes
from source import calculations


@dataclass
class Nim(calculations.Mixin):
    """The main class to implement the Nim game
    
    This class publishes the following methods. See also the calculation methods
    from the mixin in :mod:`source.calculations`.

    **Instance Methods:**
    
        :meth:`setup_heaps`
            This creates and initiates the :attr:`heaps` attribute.
        
        :meth:`set_start`
            It sets which party is to start the game.
        
        :meth:`do_move`
            It makes a move, and swaps the :attr:`activeplayer`.
        
        :meth:`get_heapstatus`
            You can get the :attr:`heaps` status (coins in heaps) in a string.
        
        :meth:`game_end`
            It tells you whether the game has ended.

    
    **Instance Attributes:**
    
    Attributes:
        heaps: This :obj:`list` contans the number of coins in the heaps. This is
            initiated in :meth:`setup_heaps` and the actual status can be
            retrieved with :meth:`get_heapstatus`.
        activeplayer: This :obj:`str` is either the "Computer" or "Player" during
            the game. Or "start" in the very beginning when starting player
            has not been selected yet.

    Args:
        error_rate: The rate percentage of mistakes the Computer is to make. By
            deafult it is 0%, i.e. the computer never makes mistakes, and the
            Player has no chance to win (unless the Player decides who to start
            and never makes mistakes). If it was set to the extreme 100%, the
            Computer would always intentionally make wrong moves, practically
            making all attempts to lose. NB, if the Player does the same, the
            Computer may still win.
            
            For testing, when moves of both parties are calculated by the
            program, the error rate of both parties can be set with the
            :obj:`typedefs.ErrorRate`. This can be used to create
            staistics of games played by differently skilled players. NB, still,
            the 1\ :sup:`st` party will be reported in the results as Computer,
            and the 2\ :sup:`nd` as Player, just to name them.
        heapcount_range: The min and max number of heaps to be set up. Less than
            3 heaps makes little sense, more than 15 is a bit of an overkill.
            So, the default is (3-15), but it can be set otherwise. NB, since
            the heaps are displayed with letters (A-Z), the heap number is
            limited to 26.
        coincount_range: The min and max number of coins to be set up in a
            heap. Less than 1 coin makes no sense, more than 20 is a bit of an
            overkill. So, the default is (1-20). NB, since the heap contents are
            displayed in max 2 digits, the coin number is limited to 99.
        misere: Whether it is the "misère" version of the game.
    """
    error_rate: ErrorRate_T=0
    heapcount_range: HeapCoinRange_T=(3, 15)
    coincount_range: HeapCoinRange_T=(1, 20)
    misere: bool=True
    
    
    def __post_init__(self):
        """Validate all input parameters for the instantiation
        
        This is quite a paranoid overkill, but it was created as an excercise of
        playing with a valid8 package.
        
        If any validation fails, exception is raised with detailed messages.
        """
        #validate the error rate
        self.set_error_rate(self.error_rate)
        
        #validate the heapcount_range type
        v8.validate(
            name='heapcount_range', 
            value=self.heapcount_range, 
            instance_of=HeapCoinRangeTypes, 
            custom=v8.validation_lib.collections.has_length(2)
        )
        #make sure the heap range is a namedtuple from here on
        self.heapcount_range = HeapCoinRange(*self.heapcount_range)
        #validate the heap range ends; less than 1 heap makes no any sense; 26
        #  heaps are from A to Z, no more to allow
        v8.validate(
            name='heapcount_range.min', 
            value=self.heapcount_range.min, 
            instance_of=int, 
            min_value=1
        )
        v8.validate(
            name='heapcount_range.max', 
            value=self.heapcount_range.max, 
            instance_of=int, 
            max_value=26
        )
        #validate the heap range ends, i.e. min <= max
        v8.validate(
            name='heapcount_range min<max', 
            value=self.heapcount_range.max-self.heapcount_range.min, 
            min_value=0, 
            help_msg='Range max bigger than min'
        )
        
        #validate the coincount_range type
        v8.validate(
            name='coincount_range', 
            value=self.coincount_range, 
            instance_of=HeapCoinRangeTypes
        )
        #validate the coincount_range size
        v8.validate(
            name='coincount_range', 
            value=self.coincount_range, 
            custom=v8.validation_lib.collections.has_length(2)
        )
        #make sure the coin range is a namedtuple from here on
        self.coincount_range = HeapCoinRange(*self.coincount_range)
        #validate the coin range ends
        v8.validate(
            name='coincount_range.min', 
            value=self.coincount_range.min, 
            instance_of=int
        )
        v8.validate(
            name='coincount_range.max', 
            value=self.coincount_range.max, 
            instance_of=int
        )
        #validate the coin range min; must be a natural number, including zero
        v8.validate(
            name='"coincount_range" min', 
            value=self.coincount_range.min, 
            min_value=0
        )
        #validate the coin range max; being displayed as 2 digit, its max is 99
        v8.validate(
            name='"coincount_range" max', 
            value=self.coincount_range.max, 
            max_value=99
        )
        #validate the coin range ends, i.e. min <= max
        v8.validate(
            name='coincount_range min<max', 
            value=self.coincount_range.max-self.coincount_range.min, 
            min_value=0, 
            help_msg='Range max bigger than min'
        )
        
        #validate the misere flag
        v8.validate(
            name='misere', 
            value=self.misere, 
            instance_of=bool
        )
        
        #reset active player
        self.activeplayer='start'
    
    
    def setup_heaps(self, 
        heapcounts: list=None
    ) -> None:
        """Initial setup of heaps of coins
        
        If `heapcounts` is provided the number of heaps and the number of coins
        in each heap is checked whether they fall into the set valid range.
        Exception is raised if not.
        
        Args:
            heapcounts: The list elements state the number of coins in each
                heap. If no list is provided, a random number of heaps are
                created with a random number of coins.
        """
        #random heaps to be created?
        if not heapcounts:
            heapcounts = []
            #create a random int for the number of heaps
            heapnumber = random.randint(*self.heapcount_range)
            for _ in range(heapnumber):
                #create a random int for the number of coins in the heap
                coincount = random.randint(*self.coincount_range)
                #create the heap with the number of coins
                heapcounts.append(coincount)
        else:
            #validate the number of heaps
            v8.validate(
                name='Heaps in "heapcounts"', 
                value=len(heapcounts), 
                custom=v8.validation_lib.comparables.between(
                    *self.heapcount_range
                )
            )
            #loop on all heaps
            for idx, heap in enumerate(heapcounts):
                #validate the number of coins in the given heap
                v8.validate(
                    name=f'Coins of heap#{idx} in "heapcounts"', 
                    value=heap, 
                    custom=v8.validation_lib.comparables.between(
                        *self.coincount_range
                    )
                )

        #create the heaps list
        self.heaps = heapcounts


    def set_start(self, 
            myturn: MyTurn_T='a'
    ) -> None:
        """Set the starting party
        
        User can explicitly set which party is to start. This makes sense if
        this decision is made based on the starting heap status.
        
        You can also ask for random selection.
        
        By default the computer makes the decision using 2 factors. First it
        figures out whether the heap status is beneficial for starting, i.e. it
        is a winning status. Then it checks the required error rate and it
        intentionally makes a wrong decision with higher likelihood if the
        error_rate is higher.
        
        The `myturn` argument is converted to a boolean, if it arrives as a
        :obj:`str`, see the parameter description below. No issues with
        "overwriting" it, because :obj:`str` is unmutable, i.e. the default
        remains the 'a' for subsequent calls and also the outer variable, where
        the argument comes from, remains unaffected.
        
        The (converted) `myturn` value is then used in a :obj:`bool` context.
        If it is of any other type, it is still interpreted as a bool. E.g. an
        empty list will act like a False, i.e. Computer will start. 
        
        The attribute `nextturn` is a :meth:`__next__` of
        :obj:`itertools.cycle`. Calling "nextturn()" infinitely yields the
        swapped player names. This attribute is used in :meth:`do_move` to
        change the active player.
        
        Args:
            myturn: The logical value from the Player's point of view.
            
                -   'a' (by default): the Computer figures out which party is to
                    start the game, based on the heap counts and the required
                    error rate
                -   truthy or 'p': Player
                -   falsy or 'c': Computer
                -   'r': random
        """
        #auto?
        if myturn=='a':
            #figure out whether the initial heap status is good for the Computer
            winning = self.is_winning_for_next()
            #whether to make a good decision
            smart = self.make_good_choice()
            
            #winning&smart or !winning&!smart makes the Computer start
            myturn = winning ^ smart
        
        #random?
        elif myturn=='r':
            myturn = random.choice([True, False])
        
        #map c to False, p to True
        elif isinstance(myturn, str):
            myturn = myturn.lower()
            if myturn=='c':
                myturn = False
            elif myturn=='p':
                myturn = True
            else:
                raise ValueError(f'"myturn" has an unknown value: "{myturn}"')
        
        #create the "nextturn" cyclic iterator function
        self.nextturn=itertools.cycle(('Computer', 'Player')).__next__
        
        #Player's turn needed?
        if myturn:
            #the cycle starts by the Computer, so get rid of it
            self.nextturn()
        #get the next player
        self.activeplayer = self.nextturn()


    def set_error_rate(self,
        error_rate: ErrorRate_T
    ) -> None:
        """Set the required error rate, even during a game

        Args:
            error_rate: The error rate to be set from the next move on 
        """
        #validate the error rate
        v8.validate(
            name='error_rate', 
            value=error_rate, 
            instance_of=ErrorRateTypes
        )
        if isinstance(error_rate, ErrorRate):
            for player in ErrorRate._fields:
                v8.validate(
                    name=f'error_rate.{player}', 
                    value=getattr(error_rate, player), 
                    instance_of=int, 
                    custom=v8.validation_lib.comparables.between(0, 100)
                )
        else:
            v8.validate(
                name='error_rate', 
                value=error_rate, 
                instance_of=int, 
                custom=v8.validation_lib.comparables.between(0, 100)
            )
        
        self.error_rate = error_rate
    
    
    def do_move(self, 
        move: Move
    ) -> None:
        """Remove given number of coins from a given heap
        
        Before doing the coin removal, check for illegal moves and report
        issues.
        
        After the given number of coins have been removed from a given heap,
        swap the active player.

        Args:
            move: The designation of the heap and the number of coins to be
                removed from that heap. Heap designation can be a letter or a
                number (starting from 0).
        """
        #heap designation letter?
        if isinstance(move.heapdesig, str):
            if len(move.heapdesig) != 1:
                raise ValueError(
                    f'Wrong heap designation ({move.heapdesig})'
                )
            heapnumber = ord(move.heapdesig.upper()) - 65
            if heapnumber < 0:
                raise ValueError(
                    f'Wrong heap designation ({move.heapdesig})'
                )
        
        #heap designation number?
        elif isinstance(move.heapdesig, int):
            heapnumber = move.heapdesig
        
        else:
            raise ValueError(
                f'Wrong heap designation type ({type(move.heapdesig)})'
            )
            
        if len(self.heaps)-1 < heapnumber:
            raise ValueError(
                f'Wrong heap letter ({chr(heapnumber+65)}), there are '
                f'A-{chr(len(self.heaps)-1+65)} heaps only'
            )

        if self.heaps[heapnumber] < move.removecount:
            raise ValueError(
                f'Heap({chr(heapnumber+65)}) only has '
                f'{self.heaps[heapnumber]} coin(s), '
                f'cannot remove {move.removecount}'
            )

        #reduce the required heap
        self.heaps[heapnumber] -= move.removecount

        #get the next player
        self.activeplayer = self.nextturn()


    def get_heapstatus(self) -> str:
        """Get the heap status in a formatted string
        
        Returns:
            The number of coins per heap. Also header at the start.
        """
        status = ''
        #at the start?
        if self.activeplayer == 'start':
            #start with the heading line
            status = ' '.join(
                [f' {chr(65+h)}' for h in range(len(self.heaps))]
            ) + '\n'
        #list the coins in the heaps, single-digit numbers padded with space
        status += ' '.join([f'{str(h):>2}' for h in self.heaps])
        
        return status
    

    def game_end(self) -> bool:
        """Identify whether the game ended, i.e. all heaps are empty

        Simply add up all the coins and return the boolean negate of the sum.
        
        Returns:
            The flag to indicate game end.
        """
        return not sum(self.heaps)



"""Play an interactive Nim Game in the CLI
"""


from source.core import Nim
from source.typedefs import Move
from source.playing.cli import ansi


def play_CLI() -> None:
    """Play one interactive game

    Create random heaps first. Then, get the user to select who to start.
    Player and Computer take turns until the game ends. Computer will make good
    or bad decisions based on the error rate set in the instance attribute.
    """
    #create instance and random heaps, then display them
    nim = Nim()
    nim.setup_heaps()
    print(nim.get_heapstatus())
    
    #by default there is no error in input
    errormsg = ''
    
    #loop on getting the Computer error rate requirement
    while True:
        errorrate = input(
            f'{errormsg}Computer error rate % [10]: '
        ) or '10'
        ansi.clean_prev_line()
        try:
            errorrate = int(errorrate)
            if not (0<=errorrate<=100):
                raise ValueError('Expecting % between 0 and 100')
            #set the error rate
            nim.set_error_rate(errorrate)
        except Exception as e:
            errormsg = str(e) + '; '
        else:
            errormsg = ''
            break
    
    #loop on getting the first move requirement
    while True:
        moveby = input(
            f'{errormsg}First move by? (Auto, Random, Computer, Player) [a]: '
        ) or 'a'
        ansi.clean_prev_line()
        try:
            #init compturn automatically
            nim.set_start(moveby)
        except ValueError as e:
            errormsg = str(e) + '; '
        else:
            errormsg = ''
            break
    
    #initiate the last move string
    lastmovestr = 'a1'
    
    #do moves in a loop, until game end
    while not nim.game_end():
        if nim.activeplayer == 'Computer':
            #make a good decision?
            if nim.make_good_choice():
                #figure out the best move
                move = nim.figure_out_best_move()
            else:
                #set a random (most probably not good) move
                move = nim.get_random_move()
        else:
            while True:
                movestr = input(
                    f'{errormsg}Heap?/Coins? [{lastmovestr}]: '
                ) or lastmovestr
                ansi.clean_prev_line()
                if not len(movestr): continue
                heapletter = movestr[:1]
                try: removecount = int(movestr[1:])
                except ValueError: continue
                if removecount < 0: continue
                move = Move(heapletter, removecount)
                lastmovestr = movestr
                break

        try:
            nim.do_move(move)
            errormsg = ''
        except ValueError as e:
            errormsg = str(e) + '; '
            continue
        print(nim.get_heapstatus())

    #The active player won after the last coin was taken by the opponent
    print(f'{nim.activeplayer} won')



"""The Nim Game calculation functions

This module defines the functions that are needed to calculate moves and
decisions in a :class:`Mixin` for the :class:`core.Nim` class.
"""


import functools
import random
from source.typedefs import Move, ErrorRate


class Mixin:
    """Class to implement the Nim game calculations
    
    This Mixin is used as an ancestor class for :class:`core.Nim`, so
    that we can define methods here, in a separate module.
    
    The term "nimsum" (Grundy value or nim-value), used here, is the Nimber_
    that the game is equivalent to. This number represents the current state of
    the game.
    
    The importance of nimsum is whether it is zero or not. If nimsum is not
    zero, it can be changed to zero by the right move of the next player.
    If the nimsum is zero, the next move will change it to non-zero, whatever
    the move is. This way, the player who moves in a way to create zero nimsum
    is the one who drives the game and can win it.
    
    In case of "misère" type of game, there is a point in the course of the game
    where the strategy must reverse, because we want to opponent to take the
    last coin. This point is the beginning of the "end-game". In the end-game,
    all moves alternate the nimsum between 0 and 1. The winning player changes
    the nimsum to 1 and the losing player has no other choise but changing it to
    zero and ultimately taking the last coin.
    
    .. _Nimber: https://en.wikipedia.org/wiki/Nimber
    """
    def make_good_choice(self) -> bool:
        """Figures out whether to make a good decision or make an error

        The required error rate indicates how bad decisions the Computer is to
        make. The higher the error rate percentage the higher the possibility
        that the Computer makes a bad move/decision. If error_rate is 0, this
        function always returns True, i.e. the Computer always makes a good
        decision. If error_rate is 100, the Computer always makes a bad
        decision.

        Returns:
            False forces to make the bad choise
        """
        #figure out the error rate for the current player
        if isinstance(self.error_rate, ErrorRate):
            activeplayer = self.activeplayer
            if activeplayer == 'start':
                activeplayer = 'Computer'
            error_rate = getattr(self.error_rate, activeplayer)
        else:
            error_rate = self.error_rate
        
        #get a random percentage
        randompercentage = random.randint(0, 99)
        
        #return true/false based on the required error rate
        return randompercentage >= error_rate


    def get_nimsum(self) -> int:
        """Calcualte the nimsum of the set of heaps

        It uses the functools.reduce() on the heaps list, with a 0 initializing
        value and uses XOR for the coin counts of the heaps.

        Returns:
            The nimsum of the heap-set
        """
        return functools.reduce(
            lambda a, b: a^b,
            self.heaps,
            0
        )


    def all_one_heaps(self) -> bool:
        """Checks whether all non-zero heaps have 1 single coin

        It uses the functools.reduce() on the heaps list, with a True
        initializing value and keep it true as long as the number of coins is at
        most one.

        Returns:
            Whether all relevant heaps have 1 single coin
        """
        return functools.reduce(
            lambda a,b: a&(b<=1),
            self.heaps,
            True
        )


    def is_winning_for_next(self) -> bool:
        """Is the current status good for the player having the next turn?

        Usually non-zero nimsum is the winning status, so that the player
        can change it to zero nimsum. This way this player can force the other
        player to make non-winning move, i.e. creating non-zero nimsum, which
        this player can change to zero nimsum again, end so on.

        However, in "misère" type game, the strategy changes in the end-game
        moves. When all heaps only have 1 coin left, the winning status is when
        nimsum is zero. I.e. the all-heaps-has-one-coin reverses the bool
        nimsum. The boolean reversal is done with an XOR.
        
        Returns:
            Whether the current status is good for the player having the next
            turn
        """
        if self.misere:
            return bool(self.get_nimsum()) ^ self.all_one_heaps()
        else:
            return bool(self.get_nimsum())


    def get_random_move(self) -> Move:
        """Figure out a random but valid move
        
        This method is used when the algorithm does not care what to move. Just
        figure out a random but valid move.
        
        Returns:
            From which heap and how many coins are to be removed in the coming
            move
        """
        #pick a random index from the active (i.e. non-empty) heaps
        heapnumber = random.choice(
            [i for i in range(len(self.heaps)) if self.heaps[i]]
        )
        #pick a random remove count from the selected random heap
        removecount = random.randint(1, self.heaps[heapnumber])
        return Move(heapnumber, removecount)


    def figure_out_best_move(self) -> Move:
        """Figure out what to move so that we get into a winning position

        First handle the situation when all heaps only have 1 coin. If so,
        there is nothing to think about, just take that 1 coin from a heap
        randomly.

        Then check whether there is only 1 heap with more than 1 coins. This is
        the indicator for the end-game reversal in the "misère" type. The
        winning strategy here is to leave odd numbers of 1-coin heaps for the
        opponent. So, if the number of heaps with 1 coin is odd, take all
        coins from the heap that has more than 1 coin. If it is even, take all
        but 1 coin. Either way, this makes sure that odd number of 1-coin
        heaps are left for the opponent.

        Otherwise, we are in the middle of the game, so figure out the all-heap
        nimsum.
        If it is zero, the player in this current turn is in a losing position,
        because whatever we do we create a non-zero nimsum that the opponent
        can change to zero nimsum again in its turn. So, having no better
        option, we just take a random number of coins from a random heap and
        hope that the opponent makes a mistake.

        If nimsum is not zero, it can be made zero by removing a cretain number
        of coins from the right heap. Check each heap and calculate the "target"
        number of coins, to null the nimsum. E.g. if the nimsum is 7 (binary
        111), we need to remove all 3 binary digits, i.e. the 4, 2 and 1,
        without changing any other binary digits, e.g. the 8. If a given heap
        contains e.g. 9 coins (binary 1001), the "target" would be the
        7 :math:`\oplus` 9::
        
            1001
            0111
            ----
            1110 => decimal 14
        
        As we cannot remove 14 coins from a heap where there are 9 coins, this
        is not a suitable heap to remove coins from. So, go on until finding a
        suitable heap. Theoretically, there must be at least 1 suitable heap.

        Returns:
            From which heap and how many coins are to be removed in the coming
            move
        """
        #active heaps have 1 coin only?
        if self.all_one_heaps():
            #empty one of the 1-coin heaps, does not matter which one
            return self.get_random_move()

        #need to handle the "only 1 heap with more than 1 coin" situation?
        if self.misere:
            #number of bigger heaps
            bigheapcount = 0
            #by default the number of 1-coin heaps is even (NB. zero is even)
            even_singlecoins = True
            #loop on all heaps
            for idx, heapcount in enumerate(self.heaps):
                #big heap?
                if heapcount>1:
                    bigheapcount += 1
                    bigheapcontent = heapcount
                    bigheapidx = idx
                
                elif heapcount==1:
                    even_singlecoins = not even_singlecoins
            
            #entering end-game, i.e. only 1 heap with more than 1 coin?
            if bigheapcount==1:
                #even number of heaps with 1 coin?
                if even_singlecoins:
                    #remove all but 1 coin from the heap which has more coins
                    removecount = bigheapcontent-1
                else:
                    #remove all coins from the heap which has more coins
                    removecount = bigheapcontent
                #remove coins so that odd number of 1-coin heaps remain
                return Move(bigheapidx, removecount)

        #calcualte the nimsum of the set of heaps
        nimsum = self.get_nimsum()

        #not a winning position, i.e. nimsum is zero?
        if not nimsum:
            #cannot do any better but initiate a random move
            return self.get_random_move()

        #look for suitable heap, so that nimsum can be nulled
        for idx, heapcount in enumerate(self.heaps):
            #calculate the wished count of this heap (i.e. so that it causes
            #zero overall nimsum)
            target_count = heapcount ^ nimsum
            #note that the required target count may be more than the heap has
            #altogether. Is this one big enough?
            if target_count < heapcount:
                removecount = heapcount - target_count
                return Move(idx, removecount) 

        #not finding a suitable heap is impossible, I must have screwed up some
        #programming
        raise Exception('Cannot find suitable heap')



"""Version in a separate module

This way it can be imported and used for package building and docs

The version info and string has the following elements:

    Major version
        to increase, when something substantial changes or not backward compatible
        anymore

    Minor version
        to increase, when new (backward compatible) features are introduced

    Patches
        to fix issues
"""

# Version of the package
__version_info__ = ('0', '1', '14')
__version__ = '.'.join(__version_info__)



"""ANSI codes for manipulating console display

These functions are used for cleaning lines after an input is taken or an error
is displayed.
"""


# An empty command seems to make the Windows cmd terminal capable of ANSI.
import os; os.system('')


ESC = '\033'
CURSOR_UP = ESC + '[A'
CURSOR_BACK = ESC + '[200D'
DELETE_LINE = ESC + '[K'


def clean_this_line():
    """Clean the input prompt inline"""
    print(
        CURSOR_BACK + DELETE_LINE, 
        end='' #supress line feed
    )


def clean_prev_line():
    """Clean the input prompt and the answer"""
    print(
        CURSOR_UP + DELETE_LINE, 
        end='' #supress line feed
    )



"""Type definitions

Users can use the following type definitions for creating and checking
arguments:

- Move: define how many coins are to be removed and from what heap
- ErrorRate: define the required error for both parties (when the program is to
  calculate the moves for both of them)
- HeapCoinRange: define heap or coin count ranges for automatic heap setup
"""


import collections
from typing import Union, List, Tuple


# A move vector, i.e. what heap is to be reduced and by how many coins
Move = collections.namedtuple('Move', 'heapdesig removecount')

# The error rate percentages can be defined for parties separately
ErrorRate = collections.namedtuple('ErrorRate', 'Computer Player')
ErrorRate_T = Union[int, ErrorRate]
ErrorRateTypes = (
    int, 
    ErrorRate
)

# How the user can indicate its requirement of the starting party
MyTurn_T = Union[str, bool]

# The accepted range of heap and coin numbers
HeapCoinRange = collections.namedtuple('HeapCoinRange', 'min max')
HeapCoinRange_T = Union[
    List[int],
    Tuple[int, int],
    HeapCoinRange
]
HeapCoinRangeTypes = (
    list, 
    tuple, 
    HeapCoinRange
)
